# Pulp ansible has the collections api broken down into a bunch of separate viewsets
# (ie: collection versions, collections, download, upload etc.) Galaxy NG expects the
# functionality of these viewsets to all be roughly the same, so instead of duplicating
# the statenents for these viewsets, they're stored here and reused on a bunch of different
# policies.
_collection_statements = [
    {
        "action": ["list", "retrieve"],
        "principal": "authenticated",
        "effect": "allow",
        "condition": "has_rh_entitlements",
    },
    {
        "action": "destroy",
        "principal": "authenticated",
        "effect": "allow",
        "condition": [
            "has_model_perms:ansible.delete_collection",
            "has_rh_entitlements",
        ],
    },
    {
        "action": ["download"],
        "principal": "authenticated",
        "effect": "allow",
        "condition": "has_rh_entitlements",
    },
    {
        "action": "create",
        "principal": "authenticated",
        "effect": "allow",
        "condition": ["can_create_collection", "has_rh_entitlements"],
    },
    {
        "action": "update",
        "principal": "authenticated",
        "effect": "allow",
        "condition": ["can_update_collection", "has_rh_entitlements"],
    },
    {
        "action": "move_content",
        "principal": "authenticated",
        "effect": "allow",
        "condition": [
            "has_model_perms:ansible.modify_ansible_repo_content",
            "has_rh_entitlements",
        ],
    },
    {
        "action": "curate",
        "principal": "authenticated",
        "effect": "allow",
        "condition": [
            "has_model_perms:ansible.modify_ansible_repo_content",
            "has_rh_entitlements",
        ],
    },
    {
        "action": "sign",
        "principal": "authenticated",
        "effect": "allow",
        "condition": [
            "can_sign_collections",
            "has_rh_entitlements"]
    },
]

_deny_all = [
    {
        "principal": "*",
        "action": "*",
        "effect": "deny"
    },
]

INSIGHTS_STATEMENTS = {
    'CollectionViewSet': _collection_statements,
    'pulp_ansible/v3/collections': _collection_statements,
    'pulp_ansible/v3/collection-versions': _collection_statements,
    'pulp_ansible/v3/collection-versions/docs': _collection_statements,
    'pulp_ansible/v3/collections/imports': _collection_statements,
    'pulp_ansible/v3/repo-metadata': _collection_statements,

    # The following endpoints are related to issue https://issues.redhat.com/browse/AAH-224
    # For now endpoints are temporary deactivated
    'pulp_ansible/v3/collection-versions/all': _deny_all,
    'pulp_ansible/v3/collections/all': _deny_all,

    # disable upload and download APIs since we're not using them yet
    'pulp_ansible/v3/collections/upload': _deny_all,
    'pulp_ansible/v3/collections/download': _deny_all,

    "NamespaceViewSet": [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_rh_entitlements",
        },
        {
            "action": "create",
            "principal": "authenticated",
            "effect": "allow",
            "condition": ["has_model_perms:galaxy.add_namespace", "has_rh_entitlements"],
        },
        {
            "action": "destroy",
            "principal": "authenticated",
            "effect": "allow",
            "condition": ["has_model_or_obj_perms:galaxy.delete_namespace", "has_rh_entitlements"],
        },
        {
            "action": "update",
            "principal": "authenticated",
            "effect": "allow",
            "condition": ["has_model_or_obj_perms:galaxy.change_namespace", "has_rh_entitlements"],
        },
    ],
    "CollectionRemoteViewSet": [
        {"action": ["list", "retrieve"], "principal": "authenticated", "effect": "allow"},
        {
            "action": ["sync", "update", "partial_update"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_model_perms:ansible.change_collectionremote",
        },
    ],
    "UserViewSet": _deny_all,
    "MyUserViewSet": [
        {
            "action": ["retrieve"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "is_current_user",
        },
    ],
    "SyncListViewSet": [
        {
            "action": ["list"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": ["has_model_perms:galaxy.view_synclist", "has_rh_entitlements"],
        },
        {
            "action": ["retrieve"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": ["has_model_or_obj_perms:galaxy.view_synclist", "has_rh_entitlements"],
        },
        {
            "action": ["destroy"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": ["has_model_perms:galaxy.delete_synclist", "has_rh_entitlements"],
        },
        {
            "action": ["create"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": ["has_model_perms:galaxy.add_synclist", "has_rh_entitlements"],
        },
        {
            "action": ["update", "partial_update"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": ["has_model_perms:galaxy.change_synclist", "has_rh_entitlements"],
        },
    ],
    "MySyncListViewSet": [
        {
            "action": ["retrieve", "list", "update", "partial_update", "curate"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": ["has_rh_entitlements", "is_org_admin"],
        },
    ],
    "TaskViewSet": [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
    ],
    "TagViewSet": [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
        },
    ],
    "GroupViewSet": [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_rh_entitlements",
        },
    ],
    "DistributionViewSet": _deny_all,
    "MyDistributionViewSet": [
        {
            "action": ["list", "retrieve"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_rh_entitlements",
        },
    ],
    "ContainerRepositoryViewSet": _deny_all,
    "LandingPageViewSet": [
        {
            "action": ["retrieve"],
            "principal": "authenticated",
            "effect": "allow",
            "condition": "has_rh_entitlements",
        },
    ],
}



from .auth import TokenView
from .sync import SyncRemoteView
from .excludes import ExcludesView
from .not_found import NotFoundView


__all__ = ("TokenView", "SyncRemoteView", "ExcludesView", "NotFoundView")



from django.db import migrations


def replace_content_guard(apps, schema_editor):
    """
    Enforce that all distributions have a content guard applied
    """
    AnsibleDistribution = apps.get_model('ansible', 'AnsibleDistribution')
    ContentRedirectContentGuard = apps.get_model('core', 'ContentRedirectContentGuard')

    OldContentGuard = apps.get_model('core', 'ContentGuard')

    # Delete the old content guard manually since it doesn't seem to get garbage collected
    # by deleting the content guard.
    OldContentGuard.objects.filter(name='ContentRedirectContentGuard').delete()

    content_guard, _ = ContentRedirectContentGuard.objects.get_or_create(
        name='ContentRedirectContentGuard',
        pulp_type='core.content_redirect'
    )

    AnsibleDistribution.objects.filter(pulp_type='ansible.ansible').update(
        content_guard=content_guard
    )


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0087_taskschedule'),
        ('galaxy', '0025_add_content_guard_to_distributions'),
    ]

    operations = [
        migrations.RunPython(
            code=replace_content_guard,
            reverse_code=migrations.RunPython.noop
        ),
    ]



"""tasks/deletion.py

This module includes tasks related to deleting Content.

You can remove Content from a Repository by making a new RepositoryVersion
without the Content. If an API endpoint uses a Distribution which points to
the latest_version of the Repository then the Content is unavailable,
however it is not deleted.

Content can only be deleted if it exists in no RepositoryVersion.

Content cannot be removed from a RepositoryVersion since it is immutable.

Pulp's orphan_cleanup task deletes any Content not part of a RepositoryVersion.

Deleting Content is made possible by retaining only one RepositoryVersion
for a Repository, and deleting all older RepositoryVersion. This module
assumes retain_repo_versions is set to 1 for all Repository the Content is
associated with.
"""

import logging

from pulp_container.app import tasks as pulp_container_tasks
from pulpcore.app.tasks import orphan_cleanup, reclaim_space
from pulpcore.plugin.tasking import general_multi_delete

log = logging.getLogger(__name__)


def delete_container_distribution(instance_ids):
    """Deletes a container distribution and push repository related."""

    log.info("Running core.general_multi_delete to delete distro and repo")
    general_multi_delete(instance_ids=instance_ids)

    log.info("Running orphan_cleanup to delete Container objects and artifacts")
    orphan_cleanup(content_pks=None, orphan_protection_time=10)


def delete_container_image_manifest(repository_pk, content_unit_pks, repo_latest_version_pk):
    """Deletes a container image manifest."""

    log.info(f"Running delete manifest for {repository_pk}")
    pulp_container_tasks.recursive_remove_content(
        repository_pk=repository_pk,
        content_units=content_unit_pks,
    )

    log.info(f"Reclaiming disk space for {repository_pk}")
    reclaim_space(repo_pks=[repository_pk], keeplist_rv_pks=[repo_latest_version_pk], force=True)



"""Utility functions for AH tests."""
import json
import logging
import os
import random
import re
import string
import tarfile
import tempfile
import time
import uuid
from contextlib import contextmanager
from subprocess import PIPE
from subprocess import run
from urllib.parse import urljoin
import shutil
import subprocess

from ansible import context
from ansible.galaxy.api import GalaxyAPI
from ansible.galaxy.api import GalaxyError
from ansible.galaxy.token import BasicAuthToken
from ansible.galaxy.token import GalaxyToken
from ansible.galaxy.token import KeycloakToken

from orionutils.generator import build_collection as _build_collection
from orionutils.generator import randstr

from .constants import USERNAME_PUBLISHER


logger = logging.getLogger(__name__)


class TaskWaitingTimeout(Exception):
    pass


class CapturingGalaxyError(Exception):
    def __init__(self, http_error, message, http_code=None):
        self.http_error = http_error
        self.message = message
        self.http_code = http_code


class CollectionInspector:

    """ Easy shim to look at tarballs or installed collections """

    def __init__(self, tarball=None, directory=None):
        self.tarball = tarball
        self.directory = directory
        self.manifest = None
        self._extract_path = None
        self._enumerate()

    def _enumerate(self):
        if self.tarball:
            self._extract_path = tempfile.mkdtemp(prefix='collection-extract-')
            cmd = f'cd {self._extract_path}; tar xzvf {self.filename}'
            subprocess.run(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
        else:
            self._extract_path = self.directory

        with open(os.path.join(self._extract_path, 'MANIFEST.json'), 'r') as f:
            self.manifest = json.loads(f.read())

        if self.tarball:
            shutil.rmtree(self._extract_path)

    @property
    def namespace(self):
        if self.manifest is None:
            return None
        return self.manifest['collection_info']['namespace']

    @property
    def name(self):
        if self.manifest is None:
            return None
        return self.manifest['collection_info']['name']

    @property
    def tags(self):
        if self.manifest is None:
            return None
        return self.manifest['collection_info']['tags']

    @property
    def version(self):
        if self.manifest is None:
            return None
        return self.manifest['collection_info']['version']


def get_client(config, require_auth=True, request_token=True, headers=None):
    """Get an API client given a role."""
    headers = headers or {}
    server = config["url"]
    assert "200" not in server
    auth_url = config.get("auth_url")

    # force the galaxy client lib to think the ignore certs kwarg was used
    # NOTE: this does not work with 2.12+
    context.CLIARGS = {"ignore_certs": True}

    # request token implies that upstream test wants to use authentication.
    # however, some tests need to auth but send request_token=False, so this
    # kwarg is poorly named and confusing.
    token = config.get("token") or None
    if request_token:
        if token:
            # keycloak must have a unique auth url ...
            if auth_url:
                token = KeycloakToken(config["token"], auth_url=auth_url)
            else:
                token = GalaxyToken(config["token"])
        else:
            token = BasicAuthToken(config["username"], config["password"])
    else:
        if require_auth:
            token = BasicAuthToken(config["username"], config["password"])
        else:
            token = None
    client = GalaxyAPI(None, "automation_hub", url=server, token=token)

    # Fix for 2.12+
    client.validate_certs = False

    # make an api call with the upstream galaxy client lib from ansible core
    def request(url, *args, **kwargs):
        url = urljoin(server, url)
        req_headers = dict(headers)
        kwargs_headers = kwargs.get('headers') or {}
        kwargs_headers_keys = list(kwargs_headers.keys())
        kwargs_headers_keys = [x.lower() for x in list(kwargs_headers.keys())]

        if isinstance(kwargs.get("args"), dict):
            kwargs["args"] = json.dumps(kwargs["args"])

        # Always send content-type
        if 'content-type' in kwargs_headers_keys:
            for k, v in headers.items():
                if k.lower() == 'content-type':
                    req_headers.pop(k, None)
        else:
            req_headers["Content-Type"] = "application/json"

        if req_headers:
            if "headers" in kwargs:
                kwargs["headers"].update(req_headers)
            else:
                kwargs["headers"] = req_headers

        return client._call_galaxy(url, *args, **kwargs)

    request.config = config
    return request


def upload_artifact(
    config, client, artifact, hash=True, no_filename=False, no_file=False, use_distribution=True
):
    """
    Publishes a collection to a Galaxy server and returns the import task URI.

    :param collection_path: The path to the collection tarball to publish.
    :return: The import task URI that contains the import results.
    """
    collection_path = artifact.filename
    with open(collection_path, "rb") as collection_tar:
        data = collection_tar.read()

    def to_bytes(s, errors=None):
        return s.encode("utf8")

    boundary = "--------------------------%s" % uuid.uuid4().hex
    file_name = os.path.basename(collection_path)
    part_boundary = b"--" + to_bytes(boundary, errors="surrogate_or_strict")

    from ansible.utils.hashing import secure_hash_s
    from ansible.galaxy.api import _urljoin

    form = []

    if hash:
        if isinstance(hash, bytes):
            b_hash = hash
        else:
            from hashlib import sha256

            b_hash = to_bytes(secure_hash_s(data, sha256), errors="surrogate_or_strict")
        form.extend([part_boundary, b'Content-Disposition: form-data; name="sha256"', b_hash])

    if not no_file:
        if no_filename:
            form.extend(
                [
                    part_boundary,
                    b'Content-Disposition: file; name="file"',
                    b"Content-Type: application/octet-stream",
                ]
            )
        else:
            form.extend(
                [
                    part_boundary,
                    b'Content-Disposition: file; name="file"; filename="%s"' % to_bytes(file_name),
                    b"Content-Type: application/octet-stream",
                ]
            )
    else:
        form.append(part_boundary)

    form.extend([b"", data, b"%s--" % part_boundary])

    data = b"\r\n".join(form)

    headers = {
        "Content-type": "multipart/form-data; boundary=%s" % boundary,
        "Content-length": len(data),
    }

    n_url = ""
    if use_distribution:
        n_url = (
            _urljoin(
                config["url"],
                "content",
                f"inbound-{artifact.namespace}",
                "v3",
                "artifacts",
                "collections",
            )
            + "/"
        )
    else:
        n_url = _urljoin(config["url"], "v3", "artifacts", "collections") + "/"

    resp = client(n_url, args=data, headers=headers, method="POST", auth_required=True)

    return resp


def wait_for_task(api_client, resp, timeout=300):
    ready = False
    url = urljoin(api_client.config["url"], resp["task"])
    wait_until = time.time() + timeout
    while not ready:
        if wait_until < time.time():
            raise TaskWaitingTimeout()
        try:
            resp = api_client(url)
        except GalaxyError as e:
            if "500" not in str(e):
                raise
        else:
            ready = resp["state"] not in ("running", "waiting")
        time.sleep(5)
    return resp


@contextmanager
def modify_artifact(artifact):
    filename = artifact.filename
    with tempfile.TemporaryDirectory() as dirpath:
        # unpack
        tf = tarfile.open(filename)
        tf.extractall(dirpath)

        try:
            yield dirpath

        finally:
            # re-pack
            tf = tarfile.open(filename, "w:gz")
            for name in os.listdir(dirpath):
                tf.add(os.path.join(dirpath, name), name)
            tf.close()


def approve_collection(client, collection):
    """Approve a collection version by moving it from the staging to published repository."""
    move_collection(client, collection, "staging", "published")


def reject_collection(client, collection):
    """Reject a collection version by moving it from the staging to published repository."""
    move_collection(client, collection, "staging", "rejected")


def move_collection(client, collection, source, destination):
    """Move a collection version between repositories.

    For use in versions of the API that implement repository-based approval.
    """
    namespace = collection.namespace
    name = collection.name
    version = collection.version
    url = f"/v3/collections/{namespace}/{name}/versions/{version}/move/{source}/{destination}"
    client(url, method="PUT")


def uuid4():
    """Return a random UUID4 as a string."""
    return str(uuid.uuid4())


def ansible_galaxy(
    command,
    check_retcode=0,
    server="automation_hub",
    ansible_config=None,
    cleanup=True
):

    tdir = tempfile.mkdtemp(prefix='ansible-galaxy-testing-')
    if not os.path.exists(tdir):
        os.makedirs(tdir)
    cfgfile = os.path.join(tdir, 'ansible.cfg')
    with open(cfgfile, 'w') as f:
        f.write('[galaxy]\n')
        f.write(f'server_list = {server}\n')
        f.write('\n')
        f.write(f'[galaxy_server.{server}]\n')
        f.write(f"url={ansible_config.get('url')}\n")
        if ansible_config.get('auth_url'):
            f.write(f"auth_url={ansible_config.get('auth_url')}\n")
        f.write('validate_certs=False\n')
        f.write(f"username={ansible_config.get('username')}\n")
        f.write(f"password={ansible_config.get('password')}\n")
        if ansible_config.get('token'):
            f.write(f"token={ansible_config.get('token')}\n")

    command_string = f"ansible-galaxy -vvv {command} --server={server} --ignore-certs"
    p = run(command_string, cwd=tdir, shell=True, stdout=PIPE, stderr=PIPE, env=os.environ)
    logger.debug(f"RUN {command_string}")
    logger.debug("STDOUT---")
    for line in p.stdout.decode("utf8").split("\n"):
        logger.debug(re.sub("(.\x08)+", "...", line))
    logger.debug("STDERR---")
    for line in p.stderr.decode("utf8").split("\n"):
        logger.debug(re.sub("(.\x08)+", "...", line))
    if check_retcode is not False:
        assert p.returncode == check_retcode, p.stderr.decode("utf8")
    if cleanup:
        shutil.rmtree(tdir)
    return p


def get_collections_namespace_path(namespace):
    """Get collections namespace path."""
    return os.path.expanduser(f"~/.ansible/collections/ansible_collections/{namespace}/")


def get_collection_full_path(namespace, collection_name):
    """Get collections full path."""
    return os.path.join(get_collections_namespace_path(namespace), collection_name)


def set_certification(client, collection):
    """Moves a collection from the `staging` to the `published` repository.

    For use in instances that use repository-based certification and that
    do not have auto-certification enabled.
    """
    if client.config["use_move_endpoint"]:
        url = (
            f"v3/collections/{collection.namespace}/{collection.name}/versions/"
            f"{collection.version}/move/staging/published/"
        )

        client(url, method="POST", args=b"{}")

        # no task url in response from above request, so can't intelligently wait.
        # so we'll just sleep for 1 second and hope the certification is done by then.
        dest_url = (
            f"v3/collections/{collection.namespace}/"
            f"{collection.name}/versions/{collection.version}/"
        )
        ready = False
        timeout = 5
        res = None
        while not ready:
            try:
                res = client(dest_url, method="GET")
                # if we aren't done publishing, GalaxyError gets thrown and we skip
                # past the below line and directly to the `except GalaxyError` line.
                ready = True
            except GalaxyError:
                time.sleep(1)
                timeout = timeout - 1
                if timeout < 0:
                    raise

        return res


def generate_namespace(exclude=None):
    """ Create a valid random namespace string """

    # This should be a list of pre-existing namespaces
    if exclude is None:
        exclude = []

    def is_valid(ns):
        """ Assert namespace meets backend requirements """
        if ns is None:
            return False
        if ns in exclude:
            return False
        if len(namespace) < 3:
            return False
        if len(namespace) > 64:
            return False
        for char in namespace:
            if char not in string.ascii_lowercase + string.digits:
                return False

        return True

    namespace = None
    while not is_valid(namespace):
        namespace = ''
        namespace += random.choice(string.ascii_lowercase)
        for x in range(0, random.choice(range(3, 63))):
            namespace += random.choice(string.ascii_lowercase + string.digits + '_')

    return namespace


def get_all_namespaces(api_client=None, api_version='v3'):
    """ Create a list of namespaces visible to the client """

    assert api_client is not None, "api_client is a required param"
    namespaces = []
    next_page = f'/api/automation-hub/{api_version}/namespaces/'
    while next_page:
        resp = api_client(next_page)
        namespaces.extend(resp['data'])
        next_page = resp.get('links', {}).get('next')
    return namespaces


def generate_unused_namespace(api_client=None, api_version='v3'):
    """ Make a random namespace string that does not exist """

    assert api_client is not None, "api_client is a required param"
    existing = get_all_namespaces(api_client=api_client, api_version=api_version)
    existing = dict((x['name'], x) for x in existing)
    return generate_namespace(exclude=list(existing.keys()))


def create_unused_namespace(api_client=None):
    """ Make a namespace for testing """

    assert api_client is not None, "api_client is a required param"
    ns = generate_unused_namespace(api_client=api_client)
    payload = {'name': ns, 'groups': []}
    api_client('/api/automation-hub/v3/namespaces/', args=payload, method='POST')
    return ns


def get_all_collections_by_repo(api_client=None):
    """ Return a dict of each repo and their collections """
    assert api_client is not None, "api_client is a required param"
    collections = {
        'staging': {},
        'published': {}
    }
    for repo in collections.keys():
        next_page = f'/api/automation-hub/_ui/v1/collection-versions/?repository={repo}'
        while next_page:
            resp = api_client(next_page)
            for _collection in resp['data']:
                key = (_collection['namespace'], _collection['name'], _collection['version'])
                collections[repo][key] = _collection
            next_page = resp.get('links', {}).get('next')
    return collections


def get_all_repository_collection_versions(api_client):
    """ Return a dict of each repo and their collection versions """

    repositories = [
        'staging',
        'published'
    ]

    collections = []
    for repo in repositories:
        next_page = f'/api/automation-hub/content/{repo}/v3/collections/'
        while next_page:
            resp = api_client(next_page)
            collections.extend(resp['data'])
            next_page = resp.get('links', {}).get('next')

    collection_versions = []
    for collection in collections:
        next_page = collection['versions_url']
        while next_page:
            resp = api_client(next_page)
            for cv in resp['data']:
                cv['namespace'] = collection['namespace']
                cv['name'] = collection['name']
                if 'staging' in cv['href']:
                    cv['repository'] = 'staging'
                else:
                    cv['repository'] = 'published'
                collection_versions.append(cv)
            next_page = resp.get('links', {}).get('next')

    rcv = dict(
        (
            (x['repository'], x['namespace'], x['name'], x['version']),
            x
        )
        for x in collection_versions
    )
    return rcv


def build_collection(
    base=None,
    config=None,
    filename=None,
    key=None,
    pre_build=None,
    extra_files=None,
    namespace=None,
    name=None,
    tags=None,
    version=None,
    dependencies=None
):

    if base is None:
        base = "skeleton"

    if config is None:
        config = {
            "namespace": None,
            "name": None,
            "tags": []
        }

    if namespace is not None:
        config['namespace'] = namespace
    else:
        config['namespace'] = USERNAME_PUBLISHER

    if name is not None:
        config['name'] = name
    else:
        config['name'] = randstr()

    if version is not None:
        config['version'] = version

    if dependencies is not None:
        config['dependencies'] = dependencies

    if tags is not None:
        config['tags'] = tags

    # workaround for cloud importer config
    if 'tools' not in config['tags']:
        config['tags'].append('tools')

    return _build_collection(
        base,
        config=config,
        filename=filename,
        key=key,
        pre_build=pre_build,
        extra_files=extra_files
    )



from django.urls import include, path
from rest_framework import routers
from pulp_ansible.app.urls import (
    v3_urls
)

from . import views, viewsets

router = routers.SimpleRouter()
router.register('namespaces', viewsets.NamespaceViewSet, basename='namespaces')

namespace_urls = [
    path("", include(router.urls)),
]

auth_urls = [
    path("auth/token/", views.TokenView.as_view(), name="auth-token"),
]

# these are included in the base router so that they only appear under `content/<distro>/v3`
sync_urls = [
    path(
        "sync/",
        views.SyncRemoteView.as_view(),
        name='sync'
    ),
    path(
        "sync/config/",
        viewsets.SyncConfigViewSet.as_view({"get": "retrieve", "put": "update"}),
        name="sync-config",
    ),
]

urlpatterns = [
    # >>>> OVERRIDDEN PULP ANSIBLE ENDPOINTS <<<<<

    # Some pulp ansible endpoints have to be overridden because we have special logic
    # that is specific to galaxy_ng's business needs. Until
    # https://github.com/pulp/pulp_ansible/issues/835 is resolved, the only solution
    # we have for this problem is to just override the viewset. Overridden API endpoints
    # should be put here. Please include comments that describe why the override is necesary.
    # Overridding an endpoint should be a measure of last resort.

    # Disable the unpaginated collection views
    # The following endpoints are related to issue https://issues.redhat.com/browse/AAH-224
    # For now endpoints are temporary deactivated
    path(
        "collections/all/",
        views.NotFoundView.as_view(),
        name="legacy-v3-metadata-collection-list",
    ),
    path(
        "collection_versions/all/",
        views.NotFoundView.as_view(),
        name="legacy-v3-metadata-collection-versions-list",
    ),

    # At the moment Automation Hub on console.redhat.com has a nonstandard configuration
    # for collection download as well as prometheus metrics that are used to track the
    # health of the service. Until https://issues.redhat.com/browse/AAH-1385 can be resolved
    # we need to continue providing this endpoint from galaxy_ng.
    path(
        "plugin/ansible/content/<path:distro_base_path>/collections/artifacts/<str:filename>",
        viewsets.CollectionArtifactDownloadView.as_view(),
        name="collection-artifact-download",
    ),

    # Overridden because the galaxy_ng endpoints only allow collections to be uploaded into
    # specific repositories.
    path(
        "artifacts/collections/",
        viewsets.CollectionUploadViewSet.as_view({"post": "create"}),
        name="collection-artifact-upload",
    ),

    # This is the same endpoint as `artifacts/collections/`. It can't be redirected because
    # redirects break on collection publish.
    path(
        "plugin/ansible/content/<path:distro_base_path>/collections/artifacts/",
        viewsets.CollectionUploadViewSet.as_view({"post": "create"}),
        name="collection-artifact-upload",
    ),

    # >>>> END OVERRIDDEN PULP ANSIBLE ENDPOINTS <<<<<

    # TODO: Endpoints that have not been moved to pulp ansible yet
    path(
        "collections/<str:namespace>/<str:name>/versions/<str:version>/move/"
        "<str:source_path>/<str:dest_path>/",
        viewsets.CollectionVersionMoveViewSet.as_view({"post": "move_content"}),
        name="collection-version-move",
    ),

    path("", include(v3_urls)),

    path("tasks/", viewsets.TaskViewSet.as_view({"get": "list"}), name="tasks-list"),
    path("tasks/<str:pk>/", viewsets.TaskViewSet.as_view({"get": "retrieve"}), name="tasks-detail"),
    path("excludes/", views.ExcludesView.as_view(), name="excludes-file"),
]



from io import StringIO
from django.contrib.auth import get_user_model
from django.core.management import call_command, CommandError
from django.test import TestCase
from unittest import mock

User = get_user_model()


class TestChangePasswordCommand(TestCase):
    def setUp(self):
        foo = User.objects.create(username="russell")
        foo.set_password("russel@llen$1")
        foo.save()
        super().setUp()

    @mock.patch(
        "django.contrib.auth.management.commands.changepassword.getpass.getpass"
    )
    def test_password_minimum_length(self, mock_getpass=None):
        def input_response(*args, **kwargs):
            return "12345"  # it requires 9 length

        mock_getpass.side_effect = input_response
        out = StringIO()
        err = StringIO()
        with self.assertRaisesMessage(
            CommandError,
            "Aborting password change for user 'russell' after 3 attempts",
        ):
            call_command("changepassword", "russell", stdout=out, stderr=err)

        self.assertIn(
            "This password is too short. It must contain at least 9 characters.",
            err.getvalue(),
        )



from django.core.management import BaseCommand
from guardian.shortcuts import assign_perm

from galaxy_ng.app.models.auth import Group

PE_GROUP_NAME = "system:partner-engineers"


class Command(BaseCommand):
    """
    This command creates or updates a partner engineering group
    with a standard set of permissions. Intended to be used for
    settings.GALAXY_DEPLOYMENT_MODE==insights.

    $ django-admin maintain-pe-group
    """

    help = "Creates/updates partner engineering group with permissions"

    def handle(self, *args, **options):
        pe_group, created = Group.objects.get_or_create(name=PE_GROUP_NAME)
        if created:
            self.stdout.write(f"Created group '{PE_GROUP_NAME}'")
        else:
            self.stdout.write(f"Group '{PE_GROUP_NAME}' already exists")

        pe_perms = [
            # groups
            "galaxy.view_group",
            "galaxy.delete_group",
            "galaxy.add_group",
            "galaxy.change_group",
            # users
            "galaxy.view_user",
            "galaxy.delete_user",
            "galaxy.add_user",
            "galaxy.change_user",
            # collections
            "ansible.modify_ansible_repo_content",
            "ansible.delete_collection",
            # namespaces
            "galaxy.add_namespace",
            "galaxy.change_namespace",
            "galaxy.upload_to_namespace",
            "galaxy.delete_namespace",
        ]

        for perm in pe_perms:
            assign_perm(perm, pe_group)

        self.stdout.write(f"Permissions assigned to '{PE_GROUP_NAME}'")



from django.conf import settings
from rest_framework.permissions import IsAuthenticatedOrReadOnly
from rest_framework.response import Response

from pulpcore.plugin.models import SigningService
from galaxy_ng.app.api import base as api_base


class FeatureFlagsView(api_base.APIView):
    permission_classes = [IsAuthenticatedOrReadOnly]

    def get(self, request, *args, **kwargs):
        flags = settings.get("GALAXY_FEATURE_FLAGS", {})
        _load_conditional_signing_flags(flags)
        return Response(flags)


def _load_conditional_signing_flags(flags):
    """Computes conditional flags for signing feature. ref: AAH-1690"""
    # kept for backwards compatibility to avoid breaking outdated UIs
    flags.setdefault(
        "collection_signing",
        bool(settings.get("GALAXY_COLLECTION_SIGNING_SERVICE"))
    )

    # Main signing feature switcher (replaces collection_signing)
    enabled = flags.setdefault(
        "signatures_enabled",
        flags["collection_signing"]
    )
    # Should UI require signature upload for enabling approval button?
    require_upload = flags.setdefault(
        "require_upload_signatures",
        enabled and bool(settings.get("GALAXY_REQUIRE_SIGNATURE_FOR_APPROVAL"))
    )

    # Is the system configured with a Signing Service to create signatures?
    def _signing_service_exists(settings):
        name = settings.get("GALAXY_COLLECTION_SIGNING_SERVICE")
        return SigningService.objects.filter(name=name).exists()

    can_create = flags.setdefault(
        "can_create_signatures",
        enabled and _signing_service_exists(settings)
    )

    # Is the system enabled to accept signature uploads?
    can_upload = flags.setdefault(
        "can_upload_signatures",
        enabled and require_upload or bool(settings.get("GALAXY_SIGNATURE_UPLOAD_ENABLED"))
    )
    # Does the system automatically sign automatically upon approval?
    flags.setdefault(
        "collection_auto_sign",
        enabled and can_create and bool(settings.get("GALAXY_AUTO_SIGN_COLLECTIONS"))
    )

    # Should UI show badges, text, signature blob, filters...
    flags.setdefault("display_signatures", enabled and can_create or can_upload)



"""test_namespace_management.py - Test related to namespaces.

See: https://issues.redhat.com/browse/AAH-1303

"""
import pytest
from ansible.errors import AnsibleError

from ..utils import get_client
from ..utils import generate_unused_namespace
from ..utils import get_all_namespaces

pytestmark = pytest.mark.qa  # noqa: F821


@pytest.mark.galaxyapi_smoke
@pytest.mark.namespace
@pytest.mark.parametrize(
    "api_version",
    [
        'v3',
        '_ui/v1'
    ]
)
def test_namespace_create_and_delete(ansible_config, api_version):
    """Tests whether a namespace can be created and deleted"""

    # http://192.168.1.119:8002/api/automation-hub/_ui/v1/namespaces/
    # http://192.168.1.119:8002/api/automation-hub/v3/namespaces/
    # {name: "testnamespace1", groups: []}

    config = ansible_config("ansible_partner")
    api_client = get_client(config, request_token=True, require_auth=True)

    new_namespace = generate_unused_namespace(api_client=api_client, api_version=api_version)
    payload = {'name': new_namespace, 'groups': []}
    resp = api_client(f'/api/automation-hub/{api_version}/namespaces/', args=payload, method='POST')
    assert resp['name'] == new_namespace

    existing2 = get_all_namespaces(api_client=api_client, api_version=api_version)
    existing2 = dict((x['name'], x) for x in existing2)
    assert new_namespace in existing2

    # This should throw an AnsibleError because the response body is an
    # empty string and can not be parsed to JSON
    try:
        resp = api_client(
            f'/api/automation-hub/{api_version}/namespaces/{new_namespace}/',
            method='DELETE'
        )
    except AnsibleError:
        pass

    existing3 = get_all_namespaces(api_client=api_client, api_version=api_version)
    existing3 = dict((x['name'], x) for x in existing3)
    assert new_namespace not in existing3



"""Tests related to collection signing.

See: https://issues.redhat.com/browse/AAH-312
"""

import logging
import time
from urllib.parse import urljoin

import pytest
from orionutils.generator import build_collection

from ..utils import (
    get_all_collections_by_repo,
    get_all_namespaces,
    get_client,
    set_certification,
)

log = logging.getLogger(__name__)

NAMESPACE = "signing"


@pytest.fixture(scope="function")
def config(ansible_config):
    return ansible_config("ansible_partner")


@pytest.fixture(scope="function")
def api_client(config):
    return get_client(config=config, request_token=True, require_auth=True)


@pytest.fixture(scope="function")
def settings(api_client):
    return api_client("/api/automation-hub/_ui/v1/settings/")


@pytest.fixture(scope="function")
def flags(api_client):
    return api_client("/api/automation-hub/_ui/v1/feature-flags/")


@pytest.fixture(scope="function", autouse=True)
def namespace(api_client):
    # ensure namespace exists
    existing = dict((x["name"], x) for x in get_all_namespaces(api_client=api_client))
    if NAMESPACE not in existing:
        payload = {"name": NAMESPACE, "groups": []}
        api_client("/api/automation-hub/v3/namespaces/", args=payload, method="POST")
        return True  # created
    return False  # not created


def import_and_wait(api_client, artifact, upload_artifact, config):
    # import and wait ...
    resp = upload_artifact(config, api_client, artifact)
    ready = False
    url = urljoin(config["url"], resp["task"])
    log.info("Waiting for collection to be imported: %s", url)
    while not ready:
        resp = api_client(url)
        ready = resp["state"] not in ("running", "waiting")
        time.sleep(1)
    assert resp["state"] == "completed"
    return resp


def sign_on_demand(api_client, signing_service, sign_url=None, **payload):
    """Sign a collection on demand calling /sign/collections/"""
    sign_url = sign_url or "/api/automation-hub/_ui/v1/collection_signing/"
    sign_payload = {"signing_service": signing_service, **payload}
    resp = api_client(sign_url, method="POST", args=sign_payload)
    log.info("Sign Task: %s", resp)
    # FIXME - pulp tasks do not seem to accept token auth, so no way to check task progress
    time.sleep(3)
    return resp


@pytest.mark.collection_signing
@pytest.mark.collection_move
@pytest.mark.standalone_only
def test_collection_auto_sign_on_approval(api_client, config, settings, flags, upload_artifact):
    """Test whether a collection is uploaded and automatically signed on approval
    when GALAXY_AUTO_SIGN_COLLECTIONS is set to true.
    """
    if not flags.get("collection_auto_sign"):
        pytest.skip("GALAXY_AUTO_SIGN_COLLECTIONS is not enabled")
    else:
        log.info("GALAXY_AUTO_SIGN_COLLECTIONS is enabled")

    can_sign = flags.get("can_create_signatures")
    if not can_sign:
        pytest.skip("GALAXY_COLLECTION_SIGNING_SERVICE is not configured")

    artifact = build_collection(
        "skeleton",
        config={
            "namespace": NAMESPACE,
            "tags": ["tools"],
        },
    )
    ckey = (artifact.namespace, artifact.name, artifact.version)

    # import and wait ...
    import_and_wait(api_client, artifact, upload_artifact, config)

    if settings.get("GALAXY_REQUIRE_CONTENT_APPROVAL"):
        # perform manual approval
        # Certify and check the response...
        cert_result = set_certification(api_client, artifact)
        assert cert_result["namespace"]["name"] == artifact.namespace
        assert cert_result["name"] == artifact.name
        assert cert_result["version"] == artifact.version
        assert cert_result["href"] is not None
        assert cert_result["metadata"]["tags"] == ["tools"]

    collections = get_all_collections_by_repo(api_client)
    assert ckey not in collections["staging"]
    assert ckey in collections["published"]

    signing_service = settings.get("GALAXY_COLLECTION_SIGNING_SERVICE")

    # Assert that the collection is signed on v3 api
    collection = api_client(
        "/api/automation-hub/content/published/v3/collections/"
        f"{artifact.namespace}/{artifact.name}/versions/{artifact.version}/"
    )
    assert len(collection["signatures"]) >= 1
    assert collection["signatures"][0]["signing_service"] == signing_service
    assert collection["signatures"][0]["signature"] is not None
    assert collection["signatures"][0]["signature"].startswith("-----BEGIN PGP SIGNATURE-----")
    assert len(collection["signatures"][0]["signature"]) >= 256
    assert collection["signatures"][0]["pubkey_fingerprint"] is not None
    assert collection["signatures"][0]["pulp_created"] is not None

    # Assert that the collection is signed on UI API
    collection_on_ui = api_client(
        "/api/automation-hub/_ui/v1/repo/published/"
        f"?deprecated=false&namespace={NAMESPACE}&name={artifact.name}"
        f"&sign_state=signed&version={artifact.version}"
    )["data"][0]
    assert collection_on_ui["sign_state"] == "signed"
    assert collection_on_ui["total_versions"] == 1
    assert collection_on_ui["signed_versions"] == 1
    assert collection_on_ui["unsigned_versions"] == 0
    metadata = collection_on_ui["latest_version"]["metadata"]
    assert len(metadata["signatures"]) >= 1
    assert metadata["signatures"][0]["signing_service"] == signing_service
    assert metadata["signatures"][0]["signature"] is not None
    assert metadata["signatures"][0]["signature"].startswith("-----BEGIN PGP SIGNATURE-----")
    assert metadata["signatures"][0]["pubkey_fingerprint"] is not None


@pytest.mark.collection_signing
@pytest.mark.standalone_only
@pytest.mark.parametrize(
    "sign_url",
    [
        "/api/automation-hub/_ui/v1/collection_signing/",
        "/api/automation-hub/_ui/v1/collection_signing/{distro_base_path}/",
        "/api/automation-hub/_ui/v1/collection_signing/{distro_base_path}/{namespace}/",
        (
            "/api/automation-hub/_ui/v1/collection_signing/"
            "{distro_base_path}/{namespace}/{collection}/"
        ),
        (
            "/api/automation-hub/_ui/v1/collection_signing/"
            "{distro_base_path}/{namespace}/{collection}/{version}/"
        ),
    ],
)
def test_collection_sign_on_demand(api_client, config, settings, flags, upload_artifact, sign_url):
    """Test whether a collection can be signed on-demand by calling _ui/v1/collection_signing/"""
    if not settings.get("GALAXY_REQUIRE_CONTENT_APPROVAL"):
        pytest.skip(
            "GALAXY_REQUIRE_CONTENT_APPROVAL is False, "
            "so content is automatically signed during approval"
        )

    can_sign = flags.get("can_create_signatures")
    if not can_sign:
        pytest.skip("GALAXY_COLLECTION_SIGNING_SERVICE is not configured")

    artifact = build_collection(
        "skeleton",
        config={
            "namespace": NAMESPACE,
            "tags": ["tools"],
        },
    )
    ckey = (artifact.namespace, artifact.name, artifact.version)

    # import and wait ...
    import_and_wait(api_client, artifact, upload_artifact, config)

    # Collection must be on /staging/
    collections = get_all_collections_by_repo(api_client)
    assert ckey in collections["staging"]
    assert ckey not in collections["published"]

    signing_service = settings.get("GALAXY_COLLECTION_SIGNING_SERVICE")

    # Sign the collection
    sign_payload = {
        "distro_base_path": "staging",
        "namespace": NAMESPACE,
        "collection": artifact.name,
        "version": artifact.version,
    }
    sign_on_demand(api_client, signing_service, sign_url.format(**sign_payload), **sign_payload)
    # Assert that the collection is signed on v3 api
    collection = api_client(
        "/api/automation-hub/content/staging/v3/collections/"
        f"{artifact.namespace}/{artifact.name}/versions/{artifact.version}/"
    )
    assert len(collection["signatures"]) >= 1
    assert collection["signatures"][0]["signing_service"] == signing_service
    assert collection["signatures"][0]["signature"] is not None
    assert collection["signatures"][0]["signature"].startswith("-----BEGIN PGP SIGNATURE-----")
    assert len(collection["signatures"][0]["signature"]) >= 256
    assert collection["signatures"][0]["pubkey_fingerprint"] is not None
    assert collection["signatures"][0]["pulp_created"] is not None

    # Assert that the collection is signed on UI API
    collection_on_ui = api_client(
        "/api/automation-hub/_ui/v1/repo/staging/"
        f"?deprecated=false&namespace={NAMESPACE}&name={artifact.name}"
        f"&sign_state=signed&version={artifact.version}"
    )["data"][0]
    assert collection_on_ui["sign_state"] == "signed"
    assert collection_on_ui["total_versions"] == 1
    assert collection_on_ui["signed_versions"] == 1
    assert collection_on_ui["unsigned_versions"] == 0
    metadata = collection_on_ui["latest_version"]["metadata"]
    assert len(metadata["signatures"]) >= 1
    assert metadata["signatures"][0]["signing_service"] == signing_service
    assert metadata["signatures"][0]["signature"] is not None
    assert metadata["signatures"][0]["signature"].startswith("-----BEGIN PGP SIGNATURE-----")
    assert metadata["signatures"][0]["pubkey_fingerprint"] is not None


@pytest.mark.collection_signing
@pytest.mark.collection_move
@pytest.mark.standalone_only
def test_collection_move_with_signatures(api_client, config, settings, flags, upload_artifact):
    """Test whether a collection can be moved from repo to repo with its
    signatures.
    """

    can_sign = flags.get("can_create_signatures")
    if not can_sign:
        pytest.skip("GALAXY_COLLECTION_SIGNING_SERVICE is not configured")

    artifact = build_collection(
        "skeleton",
        config={
            "namespace": NAMESPACE,
            "tags": ["tools"],
        },
    )
    ckey = (artifact.namespace, artifact.name, artifact.version)

    # import and wait ...
    import_and_wait(api_client, artifact, upload_artifact, config)

    # Collection must be on /staging/
    collections = get_all_collections_by_repo(api_client)
    assert ckey in collections["staging"]
    assert ckey not in collections["published"]

    signing_service = settings.get("GALAXY_COLLECTION_SIGNING_SERVICE")

    if settings.get("GALAXY_REQUIRE_CONTENT_APPROVAL"):
        # Sign the collection while still on staging
        sign_payload = {
            "distro_base_path": "staging",
            "namespace": NAMESPACE,
            "collection": artifact.name,
            "version": artifact.version,
        }
        sign_on_demand(api_client, signing_service, **sign_payload)

        # Assert that the collection is signed on v3 api
        collection = api_client(
            "/api/automation-hub/content/staging/v3/collections/"
            f"{artifact.namespace}/{artifact.name}/versions/{artifact.version}/"
        )
        assert len(collection["signatures"]) >= 1
        assert collection["signatures"][0]["signing_service"] == signing_service

        # Assert that the collection is signed on UI API
        collections = get_all_collections_by_repo(api_client)
        assert collections["staging"][ckey]["sign_state"] == "signed"

        # Move the collection to /published/
        cert_result = set_certification(api_client, artifact)
        assert cert_result["namespace"]["name"] == artifact.namespace
        assert cert_result["name"] == artifact.name
        assert cert_result["version"] == artifact.version
        assert cert_result["href"] is not None
        assert cert_result["metadata"]["tags"] == ["tools"]
        assert len(cert_result["signatures"]) >= 1

    # After moving to /published/
    # Assert that the collection is signed on v3 api
    collection = api_client(
        "/api/automation-hub/content/published/v3/collections/"
        f"{artifact.namespace}/{artifact.name}/versions/{artifact.version}/"
    )
    assert len(collection["signatures"]) >= 1
    assert collection["signatures"][0]["signing_service"] == signing_service
    assert collection["signatures"][0]["signature"] is not None
    assert collection["signatures"][0]["signature"].startswith("-----BEGIN PGP SIGNATURE-----")
    assert len(collection["signatures"][0]["signature"]) >= 256
    assert collection["signatures"][0]["pubkey_fingerprint"] is not None
    assert collection["signatures"][0]["pulp_created"] is not None

    # Assert that the collection is signed on UI API
    collection_on_ui = api_client(
        "/api/automation-hub/_ui/v1/repo/published/"
        f"?deprecated=false&namespace={NAMESPACE}&name={artifact.name}"
        f"&sign_state=signed&version={artifact.version}"
    )["data"][0]
    assert collection_on_ui["sign_state"] == "signed"
    assert collection_on_ui["total_versions"] == 1
    assert collection_on_ui["signed_versions"] == 1
    assert collection_on_ui["unsigned_versions"] == 0
    metadata = collection_on_ui["latest_version"]["metadata"]
    assert len(metadata["signatures"]) >= 1
    assert metadata["signatures"][0]["signing_service"] == signing_service
    assert metadata["signatures"][0]["signature"] is not None
    assert metadata["signatures"][0]["signature"].startswith("-----BEGIN PGP SIGNATURE-----")
    assert metadata["signatures"][0]["pubkey_fingerprint"] is not None



"""test_ldap.py - tests related to ldap authentication.

See: AAH-1593
"""
import pytest
import logging

from ..utils import get_client


log = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def config(ansible_config):
    return ansible_config("ansible_partner")


@pytest.fixture(scope="function")
def config_ldap(ansible_config):
    class AnsibleConfigLDAP(ansible_config):
        def __getitem__(self, key):
            if key == "username":
                return "professor"
            if key == "password":
                return "professor"
            return super().__getitem__(key)

    return AnsibleConfigLDAP("ansible_partner")


@pytest.fixture(scope="function")
def ldap_api_client(config_ldap):
    return get_client(
        config=config_ldap, request_token=False, require_auth=True
    )


@pytest.fixture(scope="function")
def api_client(config):
    return get_client(config=config, request_token=True, require_auth=True)


@pytest.fixture(scope="function")
def settings(api_client):
    return api_client("/api/automation-hub/_ui/v1/settings/")


@pytest.mark.standalone_only
def test_ldap_is_enabled(api_client, settings):
    """test whether ldap user can login"""
    if not settings.get("GALAXY_AUTH_LDAP_ENABLED"):
        pytest.skip("GALAXY_AUTH_LDAP_ENABLED is not enabled")

    assert (
        api_client("/api/automation-hub/_ui/v1/settings/")[
            "GALAXY_AUTH_LDAP_ENABLED"
        ]
        is True
    )


@pytest.mark.standalone_only
def test_ldap_login(ldap_api_client, settings):
    """test whether ldap user can login"""
    if not settings.get("GALAXY_AUTH_LDAP_ENABLED"):
        pytest.skip("GALAXY_AUTH_LDAP_ENABLED is not enabled")

    # This test assumes the running ldap server is the
    # testing image from: rroemhild/test-openldap
    data = ldap_api_client("/api/automation-hub/_ui/v1/me/")
    assert data["username"] == "professor"
    assert data["email"] == "professor@planetexpress.com"
    assert data["first_name"] == "Hubert"
    assert data["last_name"] == "Farnsworth"
    assert data["groups"][0]["name"] == "admin_staff"



import logging
import os

from django.conf import settings
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import NotFound

from pulpcore.plugin.access_policy import AccessPolicyFromDB

from pulp_container.app.access_policy import NamespacedAccessPolicyMixin
from pulp_container.app import models as container_models

from galaxy_ng.app import models

log = logging.getLogger(__name__)


# TODO this is a function in pulpcore that needs to get moved ot the plugin api.
# from pulpcore.plugin.util import get_view_urlpattern
def get_view_urlpattern(view):
    """
    Get a full urlpattern for a view which includes a parent urlpattern if it exists.
    E.g. for repository versions the urlpattern is just `versions` without its parent_viewset
    urlpattern.

    Args:
        view(subclass rest_framework.viewsets.GenericViewSet): The view being requested.

    Returns:
        str: a full urlpattern for a specified view/viewset
    """
    if hasattr(view, "parent_viewset") and view.parent_viewset:
        return os.path.join(view.parent_viewset.urlpattern(), view.urlpattern())
    return view.urlpattern()


class AccessPolicyBase(AccessPolicyFromDB):
    """
    This class is capable of loading access policy statements from galaxy_ng's hardcoded list of
    statements as well as from pulp's access policy database table. Priority is given to statements
    that are found in the hardcoded list of statements, so if a view name for a pulp viewset is
    found there, it will be loaded over whatever is in the database. If no viewset is found that
    matches the pulp viewset name, the statements will be loaded from the database as they would
    normally be loaded in pulp ansible.

    This class has two main functions.
    1. It is configured as the default permission class in settings.py. This means it will be used
       to load access policy definitions for all of the pulp viewsets and provides a mechanism to
       override pulp viewset access policies as well as create custom policy conditions
    2. It can be subclassed and used as a permission class for viewsets in galaxy_ng. This allows
       for custom policy conditions to be declared for specific viewsets, rather than putting them
       in the base class.
    """

    _STATEMENTS = None
    NAME = None

    @property
    def galaxy_statements(self):
        """Lazily import the galaxy_statements from the statements file."""
        if self._STATEMENTS is None:
            # import here to avoid working outside django/dynaconf settings context
            from galaxy_ng.app.access_control.statements import STANDALONE_STATEMENTS  # noqa
            from galaxy_ng.app.access_control.statements import INSIGHTS_STATEMENTS  # noqa

            self._STATEMENTS = {
                "insights": INSIGHTS_STATEMENTS,
                "standalone": STANDALONE_STATEMENTS,
            }
        return self._STATEMENTS

    def _get_statements(self):
        return self.galaxy_statements[settings.GALAXY_DEPLOYMENT_MODE]

    def get_policy_statements(self, request, view):
        statements = self._get_statements()
        if self.NAME:
            return statements.get(self.NAME, [])

        try:
            viewname = get_view_urlpattern(view)
            override_ap = statements.get(viewname, None)

            if override_ap:
                return override_ap
        except AttributeError:
            pass

        # Note: for the time being, pulp-container access policies should still be loaded from
        # the databse, because we can't override the get creation hooks like this.
        return super().get_policy_statements(request, view)

    # if not defined, defaults to parent qs of None breaking Group Detail
    def scope_queryset(self, view, qs):
        return qs

    # Define global conditions here
    def _get_rh_identity(self, request):
        if not isinstance(request.auth, dict):
            log.debug("No request rh_identity request.auth found for request %s", request)
            return False

        x_rh_identity = request.auth.get("rh_identity")
        if not x_rh_identity:
            return False

        return x_rh_identity

    def has_rh_entitlements(self, request, view, permission):

        x_rh_identity = self._get_rh_identity(request)

        if not x_rh_identity:
            log.debug(
                "No x_rh_identity found when check entitlements for request %s for view %s",
                request,
                view,
            )
            return False

        entitlements = x_rh_identity.get("entitlements", {})
        entitlement = entitlements.get(settings.RH_ENTITLEMENT_REQUIRED, {})
        return entitlement.get("is_entitled", False)

    def can_update_collection(self, request, view, permission):
        if getattr(self, "swagger_fake_view", False):
            # If OpenAPI schema is requested, don't check for update permissions
            return False
        collection = view.get_object()
        namespace = models.Namespace.objects.get(name=collection.namespace)
        return request.user.has_perm("galaxy.upload_to_namespace", namespace)

    def can_create_collection(self, request, view, permission):
        data = view._get_data(request)
        try:
            namespace = models.Namespace.objects.get(name=data["filename"].namespace)
        except models.Namespace.DoesNotExist:
            raise NotFound(_("Namespace in filename not found."))
        return request.user.has_perm("galaxy.upload_to_namespace", namespace)

    def can_sign_collections(self, request, view, permission):
        # Repository is required on the CollectionSign payload
        # Assumed that if user can modify repo they can sign everything in it
        repository = view.get_repository(request)
        can_modify_repo = request.user.has_perm('ansible.modify_ansible_repo_content', repository)

        # Payload can optionally specify a namespace to filter its contents
        # Assumed that if user has access to modify namespace they can sign its contents.
        if namespace := request.data.get('namespace'):
            try:
                namespace = models.Namespace.objects.get(name=namespace)
            except models.Namespace.DoesNotExist:
                raise NotFound(_('Namespace not found.'))
            return can_modify_repo and request.user.has_perm(
                'galaxy.upload_to_namespace', namespace
            )

        # the other filtering options are content_units and name/version
        # and falls on the same permissions as modifying the main repo
        return can_modify_repo

    def unauthenticated_collection_download_enabled(self, request, view, permission):
        return settings.GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_DOWNLOAD

    def unauthenticated_collection_access_enabled(self, request, view, action):
        return settings.GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_ACCESS


class NamespaceAccessPolicy(AccessPolicyBase):
    NAME = "NamespaceViewSet"


class CollectionAccessPolicy(AccessPolicyBase):
    NAME = "CollectionViewSet"


class CollectionRemoteAccessPolicy(AccessPolicyBase):
    NAME = "CollectionRemoteViewSet"


class UserAccessPolicy(AccessPolicyBase):
    NAME = "UserViewSet"

    def user_is_superuser(self, request, view, action):
        if getattr(self, "swagger_fake_view", False):
            # If OpenAPI schema is requested, don't check for superuser
            return False
        user = view.get_object()
        return user.is_superuser

    def is_current_user(self, request, view, action):
        if getattr(self, "swagger_fake_view", False):
            # If OpenAPI schema is requested, don't check for current user
            return False
        return request.user == view.get_object()


class MyUserAccessPolicy(AccessPolicyBase):
    NAME = "MyUserViewSet"

    def is_current_user(self, request, view, action):
        return request.user == view.get_object()


class SyncListAccessPolicy(AccessPolicyBase):
    NAME = "SyncListViewSet"


class MySyncListAccessPolicy(AccessPolicyBase):
    NAME = "MySyncListViewSet"

    def is_org_admin(self, request, view, permission):
        """Check the rhn_entitlement data to see if user is an org admin"""
        x_rh_identity = self._get_rh_identity(request)

        if not x_rh_identity:
            log.debug("No x_rh_identity found for request %s for view %s", request, view)
            return False

        identity = x_rh_identity["identity"]
        user = identity["user"]
        return user.get("is_org_admin", False)


class TagsAccessPolicy(AccessPolicyBase):
    NAME = "TagViewSet"


class TaskAccessPolicy(AccessPolicyBase):
    NAME = "TaskViewSet"


class LoginAccessPolicy(AccessPolicyBase):
    NAME = "LoginView"


class LogoutAccessPolicy(AccessPolicyBase):
    NAME = "LogoutView"


class TokenAccessPolicy(AccessPolicyBase):
    NAME = "TokenView"


class GroupAccessPolicy(AccessPolicyBase):
    NAME = "GroupViewSet"


class DistributionAccessPolicy(AccessPolicyBase):
    NAME = "DistributionViewSet"


class MyDistributionAccessPolicy(AccessPolicyBase):
    NAME = "MyDistributionViewSet"


class ContainerRepositoryAccessPolicy(AccessPolicyBase):
    NAME = "ContainerRepositoryViewSet"


class ContainerReadmeAccessPolicy(AccessPolicyBase):
    NAME = "ContainerReadmeViewset"

    def has_container_namespace_perms(self, request, view, action, permission):
        readme = view.get_object()
        return request.user.has_perm(permission) or request.user.has_perm(
            permission, readme.container.namespace
        )


class ContainerNamespaceAccessPolicy(AccessPolicyBase):
    NAME = "ContainerNamespaceViewset"


class ContainerRegistryRemoteAccessPolicy(AccessPolicyBase):
    NAME = "ContainerRegistryRemoteViewSet"


class ContainerRemoteAccessPolicy(AccessPolicyBase, NamespacedAccessPolicyMixin):
    NAME = "ContainerRemoteViewSet"

    def has_distro_permission(self, request, view, action, permission):
        class FakeView:
            def __init__(self, obj):
                self.obj = obj

            def get_object(self):
                return self.obj

        # has_container_namespace_perms

        remote = view.get_object()
        repositories = remote.repository_set.all()

        # In theory there should never be more than one repository connected to a remote, but
        # the pulp apis don't prevent you from attaching as many remotes as you want to a repo.
        for repo in repositories:
            for distro in container_models.ContainerDistribution.objects.filter(repository=repo):
                dummy_view = FakeView(distro)
                if self.has_namespace_or_obj_perms(request, dummy_view, action, permission):
                    return True

        return False


class LandingPageAccessPolicy(AccessPolicyBase):
    NAME = "LandingPageViewSet"



import json
import logging
import os
import re
import time
from unittest.mock import patch
from urllib.parse import urljoin

import pytest
from orionutils.generator import build_collection, randstr

from ..constants import USERNAME_PUBLISHER
from ..utils import (
    CapturingGalaxyError,
    get_client,
    modify_artifact,
    set_certification,
    wait_for_task,
)

logger = logging.getLogger(__name__)


class ErrorTestMode:
    """Represents a test scenario for publishing a collection."""

    def __init__(self, name, file, status, resp, no_filename=False, hash=True):
        """Configure the test scenario.

        :param name: Name for use in the test ID.
        :param file: Optionally, a file to upload. Can be just a file, or a 2-tuple of
            filename and file.
        :param status: Expected HTTP response status code from the upload attempt.
        :param resp: Expected HTTP response JSON body.
        """

        self.name = name
        self.file = file
        self.status = status
        self.resp = resp
        self.no_filename = no_filename
        self.hash = hash


def gen_name_for_invalid():
    """Generate a filename for an artifact that is not valid."""

    key = randstr(8)
    return f"{USERNAME_PUBLISHER}-invalid{key}-1.0.0.tar.gz"


@pytest.mark.stage_health
@pytest.mark.parametrize("use_distribution", [True, False])
def test_api_publish(app, ansible_config, artifact, upload_artifact, use_distribution):
    """Test the most basic, valid artifact upload via the API.

    Should successfully return a task URL to get updates of the progress,
    which should indicate a successful import.
    """

    if app.config.AUTOMATION_HUB.get("hub_use_inbound") != use_distribution:
        pytest.xfail()

    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    with patch("ansible.galaxy.api.GalaxyError", CapturingGalaxyError):
        try:
            resp = upload_artifact(config, api_client, artifact, use_distribution=use_distribution)
        except CapturingGalaxyError as capture:
            error_body = capture.http_error.read()
            logger.error("Upload failed with error response: %s", error_body)
            raise
        else:
            resp = wait_for_task(api_client, resp)
            assert resp["state"] == "completed"


@pytest.mark.skip
def test_api_publish_bad_hash(ansible_config, artifact, upload_artifact):
    """Test error responses when posting to the collections endpoint."""
    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    with pytest.raises(CapturingGalaxyError) as excinfo:
        with patch("ansible.galaxy.api.GalaxyError", CapturingGalaxyError):
            upload_artifact(config, api_client, artifact, hash=b"000000000000000")
    resp = json.loads(excinfo.value.http_error.read())

    assert excinfo.value.http_error.status == 400, excinfo.value
    assert resp["errors"]
    assert resp["errors"][0]["status"] == "400"
    assert resp["errors"][0]["code"] == "invalid"
    assert resp["errors"][0]["detail"]


@pytest.mark.stage_health
def test_api_publish_invalid_tarball(ansible_config, artifact, upload_artifact):
    """Test error responses when uploading a file that is not a tarball."""
    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    with open(artifact.filename, "wb") as f:
        f.write(randstr(1024).encode("utf8"))

    resp = upload_artifact(config, api_client, artifact)

    ready = False
    url = urljoin(config["url"], resp["task"])
    while not ready:
        resp = api_client(url)
        ready = resp["state"] not in ("running", "waiting")

    assert resp["state"] == "failed"


def test_api_publish_missing_filename(ansible_config, artifact, upload_artifact):
    """Test handling of uploads missing the filename parameter."""
    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    with pytest.raises(CapturingGalaxyError) as excinfo:
        with patch("ansible.galaxy.api.GalaxyError", CapturingGalaxyError):
            resp = upload_artifact(config, api_client, artifact, no_filename=True)
    resp = json.loads(excinfo.value.http_error.read())

    assert excinfo.value.http_error.status == 400
    assert resp["errors"]
    assert resp["errors"][0]["status"] == "400"
    assert resp["errors"][0]["source"] == {"parameter": "file"}
    assert resp["errors"][0]["code"] == "invalid"
    assert resp["errors"][0]["detail"]


@pytest.mark.stage_health
def test_api_publish_broken_manifest(ansible_config, artifact, upload_artifact):
    """Test handling of uploads missing the filename parameter."""
    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    with modify_artifact(artifact) as artifact_dir:
        manifest_path = os.path.join(artifact_dir, "MANIFEST.json")
        with open(manifest_path) as fp:
            manifest = json.load(fp)
        del manifest["collection_info"]["name"]
        with open(manifest_path, "w") as fp:
            json.dump(manifest, fp)

    resp = upload_artifact(config, api_client, artifact)
    resp = wait_for_task(api_client, resp)

    assert resp["state"] == "failed"
    assert "Invalid collection metadata. 'name' is required" in resp["error"]["description"]


INVALID_NAMES = {
    "just_wrong": lambda s: "a-wrong-name.tar.gz",
    "underscore": lambda s: s.replace("-", "_"),
    "too_long": lambda s: s.replace(s.split('-')[1], 'nevergonnagiveyouuporhurtyo' + 'u' * 100),
}


@pytest.mark.parametrize("wrong_name", INVALID_NAMES)
def test_api_publish_invalid_filename(ansible_config, artifact, upload_artifact, wrong_name):
    """Test handling of uploads with invalid filenames."""
    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    # use the param lambda function to alter the tarball filename ...
    wrong_name = INVALID_NAMES[wrong_name](os.path.basename(artifact.filename))
    filename = os.path.join(os.path.dirname(artifact.filename), wrong_name)
    os.rename(artifact.filename, filename)

    # need to rename so that upload_artifact() has the correct name.
    artifact.filename = filename

    # Ensure an excepton is thrown by the client lib ...
    with pytest.raises(CapturingGalaxyError) as excinfo:
        with patch("ansible.galaxy.api.GalaxyError", CapturingGalaxyError):
            resp = upload_artifact(config, api_client, artifact)
    text = excinfo.value.http_error.read()
    assert excinfo.value.http_error.status == 400, f"{excinfo.value.http_error.status}: {text}"

    resp = json.loads(text)
    assert resp["errors"]
    assert resp["errors"][0]["status"] == "400"
    assert resp["errors"][0]["source"] == {"parameter": "filename"}
    assert resp["errors"][0]["detail"]
    assert resp["errors"][0]["code"] == "invalid"


def test_api_publish_missing_file(ansible_config, artifact, upload_artifact):
    """Test handling of POSTs to the artifact endpoint neglecting to submit a file."""
    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    with pytest.raises(CapturingGalaxyError) as excinfo:
        with patch("ansible.galaxy.api.GalaxyError", CapturingGalaxyError):
            resp = upload_artifact(config, api_client, artifact, no_file=True)
    resp = json.loads(excinfo.value.http_error.read())

    assert excinfo.value.http_error.status == 400
    assert resp["errors"]
    assert resp["errors"][0]["status"] == "400"
    assert resp["errors"][0]["source"] == {"parameter": "file"}
    assert resp["errors"][0]["code"] == "required"
    assert resp["errors"][0]["detail"]


MAX_LENGTH_AUTHOR = 64
MAX_LENGTH_LICENSE = 32
MAX_LENGTH_NAME = 64
MAX_LENGTH_TAG = 64
MAX_LENGTH_URL = 2000
MAX_LENGTH_VERSION = 128


@pytest.mark.parametrize(
    "field",
    [
        ("authors", ["name" * (MAX_LENGTH_AUTHOR + 1)], MAX_LENGTH_AUTHOR),
        ("license", "G" * (MAX_LENGTH_LICENSE + 1), MAX_LENGTH_LICENSE),
        # ("name", "n" * (MAX_LENGTH_NAME + 1), MAX_LENGTH_NAME),
        # ("namespace", "n" * (MAX_LENGTH_NAME + 1), MAX_LENGTH_NAME),
        ("tags", ["application", "x" * (MAX_LENGTH_TAG + 1)], MAX_LENGTH_TAG),
        ("repository", "http://" + "h" * MAX_LENGTH_URL, MAX_LENGTH_URL),
        ("homepage", "http://" + "h" * MAX_LENGTH_URL, MAX_LENGTH_URL),
        ("issues", "http://" + "h" * MAX_LENGTH_URL, MAX_LENGTH_URL),
        ("documentation", "http://" + "h" * MAX_LENGTH_URL, MAX_LENGTH_URL),
        # ("version", "1.1." + "1" * MAX_LENGTH_VERSION, MAX_LENGTH_VERSION),
    ],
    ids=lambda _: _[0],
)
@pytest.mark.stage_health
def test_long_field_values(ansible_config, upload_artifact, field):
    """Test handling of POSTs to the artifact endpoint neglecting to submit a file."""
    config = ansible_config("ansible_partner")
    api_client = get_client(config)
    fieldname, fieldvalue, fieldmax = field
    artifact = build_collection(
        "skeleton", config={"namespace": USERNAME_PUBLISHER, fieldname: fieldvalue}
    )

    resp = upload_artifact(config, api_client, artifact)
    resp = wait_for_task(api_client, resp)

    assert resp["state"] == "failed"

    # Should END with an error
    assert "must not be greater than %s characters" % fieldmax in resp["error"]["description"]
    assert fieldname in resp["error"]["description"]


@pytest.mark.parametrize(
    "spec",
    [
        # ("2eq", "==2.10", "completed"),
        # ("range1", ">=2.10,<2.11", "completed"),
        ("range2", ">=2.10,<<2.11", "failed"),
        ("2comma", ">=2.10,,", "failed"),
        # ("star", "2.10.*", "completed"),
        # ("exc", ">=2.1,!=2.1.2", "completed"),
        # Known limitations
        # ("label", "2.10.post1", "completed"),
        # ("pre", "2.11b4", "completed"),
        # ("1eq", "=2.10", "failed"),
        # ("contradiction", ">2.0,<1.0", "failed"),
    ],
    ids=lambda _: _[0],
)
def test_ansible_requires(ansible_config, upload_artifact, spec):
    """
    Test handling of POSTs to the artifact endpoint neglecting to submit a file.

    Also verifies that the collections endpoint properly returns a `requires_ansible` field,
    and that the returned field matches the collection metadata.
    """
    config = ansible_config("ansible_partner")
    api_client = get_client(config)
    _, requires_ansible, result = spec
    artifact = build_collection(
        "skeleton",
        config={"namespace": USERNAME_PUBLISHER},
        extra_files={"meta/runtime.yml": {"requires_ansible": requires_ansible}},
    )

    resp = upload_artifact(config, api_client, artifact)
    resp = wait_for_task(api_client, resp)

    assert resp["state"] == result

    if result == "completed":
        admin = get_client(ansible_config("ansible_insights"))
        set_certification(admin, artifact)

        collection_url = f"v3/collections/{artifact.namespace}/{artifact.name}/versions/1.0.0/"
        collection_resp = api_client(collection_url)
        assert collection_resp["requires_ansible"] == requires_ansible

        ui_collection_url = (
            f"_ui/v1/repo/published/{artifact.namespace}/{artifact.name}/?versions=1.0.0"
        )
        ui_collection_resp = api_client(ui_collection_url)
        assert ui_collection_resp["latest_version"]["requires_ansible"] == requires_ansible


@pytest.mark.stage_health
def test_ansible_lint_exception(ansible_config, upload_artifact):
    """
    Ensure that:
        * ansible-lint runs against our uploaded collection
        * the bug in https://github.com/ansible/galaxy-importer/pull/115 remains fixed.
    """
    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    broken_role_yaml = [{"name": "a task", "not.a.real.module": {"fake": "fake"}}]

    artifact = build_collection(
        "skeleton",
        config={
            "namespace": USERNAME_PUBLISHER,
            "tags": ["database"],
        },
        extra_files={
            "roles/main/tasks/main.yml": broken_role_yaml,
            "roles/main/README.md": "role readme",
        },
    )

    resp = upload_artifact(config, api_client, artifact)

    ready = False
    url = urljoin(config["url"], resp["task"])
    while not ready:
        resp = api_client(url)
        ready = resp["state"] not in ("running", "waiting")
        time.sleep(5)

    log_messages = [item["message"] for item in resp["messages"]]

    linting_re = re.compile("Linting role .* via ansible-lint")
    critical_re = re.compile("CRITICAL Couldn't parse task at")
    linting = [item for item in log_messages if linting_re.match(item)]
    critical = [item for item in log_messages if critical_re.match(item)]

    assert len(linting) == 1  # linting occurred
    assert len(critical) == 0  # no critical errors


def test_api_publish_log_missing_ee_deps(ansible_config, upload_artifact):
    """
    Test that galaxy-importer logs when meta/execution-environment.yml
    lists a python deps file or system deps file and the listed file is not found.

    In this case a requirements.txt file exists but bindep.txt does not.
    """

    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    artifact = build_collection(
        "skeleton",
        config={
            "namespace": USERNAME_PUBLISHER,
            "tags": ["cloud"],
        },
        extra_files={
            "meta/runtime.yml": {"requires_ansible": ">=2.10,<2.11"},
            "requirements.txt": ["requests  # my pypi requirement"],
            "meta/execution-environment.yml": {
                "version": "1",
                "dependencies": {"python": "requirements.txt", "system": "bindep.txt"},
            },
        },
    )

    resp = upload_artifact(config, api_client, artifact)

    ready = False
    url = urljoin(config["url"], resp["task"])
    while not ready:
        resp = api_client(url)
        ready = resp["state"] not in ("running", "waiting")
        time.sleep(5)

    log_messages = [item["message"] for item in resp["messages"]]

    file_not_found_re = re.compile(
        r"^Error when checking meta\/execution-environment.yml for dependency files: "
        r"\[Errno 2\] No such file or directory: '\/tmp\S+bindep.txt'$"
    )
    file_not_found = [item for item in log_messages if file_not_found_re.match(item)]

    assert len(file_not_found) == 1


def test_api_publish_ignore_files_logged(ansible_config, upload_artifact):
    """
    Test that galaxy-importer logs when ansible-test sanity ignore files are present.
    """
    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    artifact = build_collection(
        "skeleton",
        config={
            "namespace": USERNAME_PUBLISHER,
            "tags": ["networking"],
        },
        extra_files={
            "meta/runtime.yml": {"requires_ansible": ">=2.10,<2.11"},
            "tests/sanity/ignore-2.10.txt": [
                "plugins/action/ios.py action-plugin-docs",
                "plugins/modules/ios_vlan.py validate-modules:deprecation-mismatch",
            ],
        },
    )

    resp = upload_artifact(config, api_client, artifact)

    ready = False
    url = urljoin(config["url"], resp["task"])
    while not ready:
        resp = api_client(url)
        ready = resp["state"] not in ("running", "waiting")
        time.sleep(5)

    log_messages = [item["message"] for item in resp["messages"]]

    ignorefile_re = re.compile(
        "Ignore files skip ansible-test sanity tests, found ignore-2.10.txt with 2 statement"
    )
    ignorefile = [item for item in log_messages if ignorefile_re.match(item)]

    assert len(ignorefile) == 1  # found ignorefile log message


@pytest.mark.cloud_only
def test_publish_fail_required_tag(ansible_config, upload_artifact):
    """
    Test cloud publish fails when collection metadata tags do not include
    at least one tag in the galaxy-importer REQUIRED_TAG_LIST,
    as set by the galaxy-importer config CHECK_REQUIRED_TAGS.
    """
    config = ansible_config("ansible_partner")
    api_client = get_client(config)

    artifact = build_collection(
        "skeleton",
        config={
            "namespace": USERNAME_PUBLISHER,
            "tags": ["not_a_required_tag"],
        },
    )

    resp = upload_artifact(config, api_client, artifact)

    ready = False
    url = urljoin(config["url"], resp["task"])
    while not ready:
        resp = api_client(url)
        ready = resp["state"] not in ("running", "waiting")
        time.sleep(5)

    assert resp["state"] == "failed"
    assert (
        "Invalid collection metadata. At least one tag required from tag list"
        in resp["error"]["description"]
    )



"""test_synclist.py - Test related to synclists."""

import os
import time
from contextlib import contextmanager

import pytest
from ansible.galaxy.api import GalaxyError
from orionutils.generator import build_collection

from ..constants import USERNAME_PUBLISHER
from ..utils import get_client, set_certification, wait_for_task

pytestmark = pytest.mark.qa  # noqa: F821


# TODO: see if ansible_config() does or can be made to accept a user param
@contextmanager
def set_username_env_var(hub_username):
    try:
        original_username = os.environ.get("HUB_USERNAME", "")
        os.environ["HUB_USERNAME"] = hub_username
        yield
    finally:
        os.environ["HUB_USERNAME"] = original_username


@pytest.mark.galaxyapi_smoke
@pytest.mark.synclist
@pytest.mark.cloud_only
def test_synclist_object_get(ansible_config):
    """Validate user can read the endpoints with synclist objects."""

    # set user to org-admin to access synclists
    with set_username_env_var("org-admin"):
        config = ansible_config("ansible_partner")
        api_client = get_client(config, request_token=True, require_auth=True)

    resp = api_client("_ui/v1/my-synclists/", args={}, method="GET")
    assert resp["meta"]["count"] == 1
    print(f"resp={resp}")

    resp = api_client("_ui/v1/synclists/", args={}, method="GET")
    assert resp["meta"]["count"] > 0


@pytest.mark.galaxyapi_smoke
@pytest.mark.synclist
@pytest.mark.cloud_only
def test_synclist_object_edit(ansible_config, upload_artifact):
    """Edit the synclist json object and confirm data is saved.

    Note: the synclist object is not updated by adding one or removing one
    object like a collection, the json object needs to be updated to its
    future state before sending the PUT, and the synclist is overwritten.
    """

    # {
    #     'id': 2,
    #     'name': '6089720-synclist',
    #     'policy': 'exclude',
    #     'upstream_repository': '8b7f16b8-9753-4389-b6f6-32d5a8cbe247',
    #     'repository': 'b5dcd5b4-787b-4c59-84aa-96e87adc2014',
    #     'collections': [],
    #     'namespaces': [],
    #     'groups': [
    #         {
    #             'id': 2,
    #             'name': 'rh-identity-account:6089720',
    #             'object_permissions': [
    #                 'view_synclist', 'add_synclist', 'delete_synclist', 'change_synclist'
    #             ]
    #         }
    #     ]
    # }

    # set user to org-admin to access synclists
    with set_username_env_var("org-admin"):
        config = ansible_config("ansible_partner")
        api_client = get_client(config, request_token=True, require_auth=True)

    # determine synclist repo associated to user
    resp = api_client("_ui/v1/my-synclists/", args={}, method="GET")
    synclist_data_before = resp["data"][0]
    synclist_id = synclist_data_before["id"]

    # edit synclist payload
    my_synclist_url = f"_ui/v1/my-synclists/{synclist_id}/"
    synclist_data_after = dict(synclist_data_before)
    synclist_data_after["namespaces"] = [USERNAME_PUBLISHER]
    resp = api_client(my_synclist_url, args=synclist_data_after, method="PUT")

    # confirm synclist GET payload is same as payload sent via PUT
    resp = api_client(my_synclist_url, args={}, method="GET")
    assert resp == synclist_data_after

    # return synclist to previous state
    api_client(my_synclist_url, args=synclist_data_before, method="PUT")
    resp = api_client(my_synclist_url, args={}, method="GET")
    assert resp == synclist_data_before


@pytest.mark.galaxyapi_smoke
@pytest.mark.synclist
@pytest.mark.cloud_only
def test_synclist_object_and_synclist_repo_edit(ansible_config, upload_artifact):
    """Edit synclist object to exclude a collection, do curate task,
    confirm collection removed from synclist repository."""

    # NOTE: on stage env, a toggle action accesses these:
    # PUT https://console.stage.redhat.com/api/automation-hub/_ui/v1/my-synclists/1/
    # POST https://console.stage.redhat.com/api/automation-hub/_ui/v1/my-synclists/1/curate/

    config = ansible_config("ansible_partner")
    api_client = get_client(config, request_token=True, require_auth=True)

    # create and certify a new collection
    collection = build_collection("skeleton", config={"namespace": USERNAME_PUBLISHER})
    resp = upload_artifact(config, api_client, collection)
    resp = wait_for_task(api_client, resp)
    set_certification(api_client, collection)
    collection_key = (collection.namespace, collection.name)

    # set user to org-admin to access synclists
    with set_username_env_var("org-admin"):
        config = ansible_config("ansible_partner")
        api_client = get_client(config, request_token=True, require_auth=True)

    # determine synclist repo associated to user
    resp = api_client("_ui/v1/my-synclists/", args={}, method="GET")
    synclist_data_before = resp["data"][0]
    synclist_name = synclist_data_before["name"]
    synclist_id = synclist_data_before["id"]

    # check that collection is part of synclist repo
    url = f"content/{synclist_name}/v3/collections/?limit=30"
    resp = api_client(url)
    collections_before = [(c["namespace"], c["name"]) for c in resp["data"]]
    assert collection_key in collections_before

    # edit synclist payload
    my_synclist_url = f"_ui/v1/my-synclists/{synclist_id}/"
    synclist_data_after = dict(synclist_data_before)
    synclist_data_after["collections"] = [
        {"namespace": collection.namespace, "name": collection.name}
    ]
    resp = api_client(my_synclist_url, args=synclist_data_after, method="PUT")

    # kick off a curate task
    resp = api_client(f"_ui/v1/my-synclists/{synclist_id}/curate/", args={}, method="POST")

    # wait for the curate task to finish
    try:
        wait_for_task(api_client, resp)
    except GalaxyError as ge:
        # FIXME - pulp tasks do not seem to accept token auth
        if ge.http_code in [403, 404]:
            time.sleep(5)
        else:
            raise Exception(ge)

    # check collection is NOT part of synclist repo
    url = f"content/{synclist_name}/v3/collections/?limit=30"
    resp = api_client(url)
    collections_after = [(c["namespace"], c["name"]) for c in resp["data"]]
    assert collection_key not in collections_after



"""test_auth.py - Test related to authentication.

See: https://github.com/ansible/galaxy-dev/issues/149

"""
import pytest
from ansible.galaxy.api import GalaxyError

from ..utils import get_client
from ..utils import uuid4

pytestmark = pytest.mark.qa  # noqa: F821


@pytest.mark.skip(reason="fails in ephemeral")
@pytest.mark.parametrize("user", ("ansible_user", "ansible_partner", "ansible_insights"))
@pytest.mark.galaxyapi_smoke
def test_token_auth(user, ansible_config):
    """Test whether normal auth is required and works to access APIs.

    Also tests the settings for all three roles used for testing.
    """

    config = ansible_config("ansible_partner")

    client = get_client(config, request_token=False, require_auth=False)
    with pytest.raises(GalaxyError) as ctx:
        client("v3/collections/", method="GET")
    assert ctx.value.http_code == 403

    client = get_client(config, request_token=True)
    resp = client("", method="GET")
    assert "available_versions" in resp


@pytest.mark.skip(reason="fails in ephemeral")
@pytest.mark.galaxyapi_smoke
def test_auth_admin(ansible_config):
    """Test whether admin can not access API root using invalid token."""

    config = ansible_config("ansible_insights")
    client = get_client(
        config,
        request_token=False,
        headers={"Authorization": f"Bearer {uuid4()}"}
    )
    with pytest.raises(GalaxyError) as ctx:
        # url not provided defaults to API root.
        client("", method="GET")
    assert ctx.value.http_code == 403


@pytest.mark.skip(reason="fails in ephemeral")
@pytest.mark.galaxyapi_smoke
def test_auth_exception(ansible_config, published):
    """Test whether an HTTP exception when using an invalid token."""

    config = ansible_config("ansible_insights")
    client = get_client(
        config,
        request_token=False,
        headers={"Authorization": f"Bearer {uuid4()}"}
    )
    with pytest.raises(GalaxyError) as ctx:
        client("", method="GET")
    assert ctx.value.http_code == 403
